import functools
import json
from pathlib import Path
from subprocess import run

import pytablewriter as ptw
from pytablereader import MarkdownTableFileLoader
from pytablewriter import MarkdownTableWriter
from pytablewriter.style import Style


def read_md(f):
    loader = MarkdownTableFileLoader(f)
    for table_data in loader.load():
        return ptw.dumps_tabledata(table_data)


def write_md(results):
    initial_message = (
        "Having a bit of fun trying to solve the problems from "
        "https://projecteuler.net/archives using different programming "
        "languages for learning purposes.\n\n"
    )
    with open("../README.md", "w") as f:
        f.write(initial_message)
        results.stream = f
        results.write_table()


def cache_and_save_results(func):
    @functools.wraps(func)
    @functools.lru_cache(maxsize=None)
    def wrapper(script_number):
        cache_file_name = f"{func.__name__.split('_')[1]}.json"
        num = int(script_number[-6:-3])

        if Path(cache_file_name).exists():
            with open(cache_file_name, "r") as f:
                cache = json.load(f)
        else:
            cache = {}

        if str(num) not in cache:
            try:
                tim, res = func(script_number)
                result = f"Result is {res.split(': ')[1]} and took {tim.split(': ')[1]}"
            except IndexError:
                tim = "Took: --"
                res = "Result is: --"
                result = "Result is -- and took --"
            cache[str(num)] = result
            with open(cache_file_name, "w") as f:
                json.dump(cache, f, indent=2, ensure_ascii=False)
        else:
            result = cache[str(num)]
            tim = f"Took: {' '.join(result.split(' ')[-2:])}"
            res = f"Result is: {result.split(' ')[2]}"

        return tim, res

    return wrapper


def get_problems(path, ext=None):
    scripts = sorted(
        [script.as_posix() for script in path.rglob(f"*[0-9].{ext}")],
        key=lambda x: x[-5:],
    )

    return scripts


@cache_and_save_results
def timing_Python(script_number):
    _run = run(
        ["python3", f"{script_number}"], capture_output=True, text=True
    ).stdout.split("\n")

    return _run[0], _run[2]


@cache_and_save_results
def timing_Julia(script_number):
    _run = run(
        ["julia", f"{script_number}"], capture_output=True, text=True
    ).stdout.split("\n")

    tim = " ".join([_run[0], " ".join(_run[1].split(" ")[2:4])])

    return tim, _run[3]


def execute(combine_columns=True, nproblems=None):
    """
    Execute scripts and return md formatted table
    """
    python_path = Path.cwd() / "Python"
    pscripts = get_problems(python_path, ext="py")

    julia_path = Path.cwd() / "Julia"
    jscripts = get_problems(julia_path, ext="jl")

    ptimings, presults = [], []
    jtimings, jresults = [], []
    for problem, _ in enumerate(pscripts[:nproblems]):
        try:
            ptim, pres = timing_Python(pscripts[problem])
            ptimings.append(ptim)
            presults.append(pres)
        except KeyError:
            ptimings.append("Took: --")
            presults.append(f"Result for problem {problem:0>3}: --")
        try:
            jtim, jres = timing_Julia(jscripts[problem])
            jtimings.append(jtim)
            jresults.append(jres)
        except IndexError:
            jtimings.append("Took: --")
            jresults.append(f"Result for problem {problem:0>3}: --")

    not_equal = []
    for npb, (pres, jres) in enumerate(zip(presults, jresults), start=1):
        if not pres.split(" ")[-1] == jres.split(" ")[-1]:
            not_equal.append([npb, pres.split(" ")[-1], jres.split(" ")[-1]])

    if combine_columns:
        headers = [
            "Problem #",
            "Result",
            "T_exec (Python)",
            "T_exec (Julia)",
        ]
        type_hints = ["str", "int", "str", "str"]
        value_matrix = [
            [
                f"{idx:0>3}",
                int(r_python.split(" ")[-1]),
                t_python.split(": ")[1],
                t_julia.split(": ")[1],
            ]
            for idx, (r_python, t_python, t_julia) in enumerate(
                zip(presults, ptimings, jtimings), start=1
            )
        ]
        discrepancies = [
            f"  - Problem {npb}: Python result is {py_r} and in Julia is {jl_r}"
            for npb, py_r, jl_r in not_equal
        ]
        writer_comb = MarkdownTableWriter(
            table_name="\nDiscrepancies in:\n" + "\n".join(discrepancies) + "\n",
            type_hints=type_hints,
            headers=headers,
            value_matrix=value_matrix,
            column_styles=[
                Style(align="center"),
                Style(align="right", thousand_separator="_"),
                Style(align="right"),
                Style(align="right"),
            ],  # specify styles for each column
            margin=1,  # add a whitespace for both sides of each cell
        )
        writer_comb.set_indent_level(3)

        return writer_comb

    else:
        languages = ["Python", "Julia"]

        for lang in languages:
            if lang == "Python":
                headers = ["Problem #", "Result", "T_exec (Python)"]
                type_hints = ["str", "int", "str"]
                value_matrix = [
                    [
                        f"{idx:0>3}",
                        int(r_python.split(" ")[-1]),
                        t_python.split(": ")[1],
                    ]
                    for idx, (r_python, t_python) in enumerate(
                        zip(presults, ptimings), start=1
                    )  # noqa: E501
                ]
            elif lang == "Julia":
                headers = ["Problem #", "Result", "T_exec (Julia)"]
                type_hints = ["str", "int", "str"]
                value_matrix = [
                    [
                        f"{idx:0>3}",
                        int(r_julia.split(" ")[-1]),
                        t_julia.split(": ")[1],
                    ]
                    for idx, (r_julia, t_julia) in enumerate(
                        zip(jresults, jtimings), start=1
                    )  # noqa: E501
                ]

            writer = MarkdownTableWriter(
                table_name=f"Results for {lang}",
                headers=headers,
                type_hints=type_hints,
                value_matrix=value_matrix,
                column_styles=[
                    Style(align="center"),
                    Style(align="right", thousand_separator="_"),
                    Style(align="right"),
                ],  # specify styles for each column
                margin=1,  # add a whitespace for both sides of each cell
            )

            return writer


if __name__ == "__main__":
    # md = read_md("../README.md")
    # print(md)
    res = execute(combine_columns=True, nproblems=-1)
    write_md(res)
